@@ -617,6 +617,8 @@ pub mod client;
 pub mod comm;
 #[cfg(feature = "cluster")]
 pub mod replication;
+#[cfg(feature = "server")]
+pub mod passwords;
 
 #[cfg(test)]
 mod test {
