@@ -1,7 +1,6 @@
 #[macro_use]
 extern crate lazy_static;
 
-use tokio::fs;
 use tokio::io::AsyncReadExt;
 use tokio::net::{TcpListener, TcpStream, UdpSocket};
 use tokio::signal::unix::{signal, SignalKind};
@@ -59,7 +58,7 @@ use serde::{Deserialize, Serialize};
 
 const ERR_INVALID_DATA_BLOCK: &str = "Invalid data block";
 const MAX_AUTH_FRAME_SIZE: usize = 1024;
-const UDP_BUF_LEN: usize = 65500;
+const DEFAULT_UDP_FRAME_SIZE: u16 = 4096;
 
 static ALLOW_ANONYMOUS: atomic::AtomicBool = atomic::AtomicBool::new(false);
 static MAX_PUB_SIZE: atomic::AtomicUsize = atomic::AtomicUsize::new(0);
@@ -116,7 +115,7 @@ macro_rules! format_login {
 
 lazy_static! {
     static ref DB: RwLock<ServerClientDB> = RwLock::new(<_>::default());
-    static ref PASSWORD_FILE: RwLock<Option<String>> = RwLock::new(None);
+    static ref PASSWORD_DB: RwLock<psrt::passwords::Passwords> = RwLock::new(<_>::default());
     static ref STATS_COUNTERS: RwLock<Counters> = RwLock::new(Counters::new());
     static ref PID_FILE: Mutex<Option<String>> = Mutex::new(None);
     static ref HOST_NAME: RwLock<String> = RwLock::new("unknown".to_owned());
@@ -531,16 +530,9 @@ async fn init_stream(
 /// # Errors
 ///
 /// Will return Err if no password fine defined or unable to read
-pub async fn authenticate(login: &str, password: &str) -> Result<bool, Error> {
-    if let Some(pfile) = PASSWORD_FILE.read().await.as_ref() {
-        let data = fs::read_to_string(pfile)
-            .await
-            .map_err(|e| Error::io(format!("Unable to read password file {}: {}", pfile, e)))?;
-        let htpasswd = htpasswd_verify::load(&data);
-        Ok(htpasswd.check(login, password))
-    } else {
-        Err(Error::access("No password file defined"))
-    }
+#[inline]
+pub async fn authenticate(login: &str, password: &str) -> bool {
+    PASSWORD_DB.read().await.verify(login, password)
 }
 
 #[inline]
@@ -581,7 +573,7 @@ async fn handle_stream(
                 addr
             )));
         }
-    } else if authenticate(login, password).await? {
+    } else if authenticate(login, password).await {
         trace!("User {} logged in from {}", login, addr);
     } else {
         trace!("Access denied for {} from {}", login, addr);
@@ -742,6 +734,7 @@ struct ConfigAuth {
 struct ConfigProto {
     bind: String,
     bind_udp: Option<String>,
+    udp_frame_size: Option<u16>,
     timeout: f64,
     tls_cert: Option<String>,
     tls_key: Option<String>,
@@ -860,41 +853,44 @@ async fn run_server(
         warn!("cluster feature is disabled");
     }
     if let Some(ref bind_udp) = config.proto.bind_udp {
-        info!("binding UDP socket to: {}", bind_udp);
+        let udp_frame_size = config
+            .proto
+            .udp_frame_size
+            .unwrap_or(DEFAULT_UDP_FRAME_SIZE);
+        info!(
+            "binding UDP socket to: {}, max frame size: {}",
+            bind_udp, udp_frame_size
+        );
         let udp_sock = UdpSocket::bind(bind_udp).await?;
         tokio::spawn(async move {
-            let mut buf: [u8; UDP_BUF_LEN] = [0; UDP_BUF_LEN];
+            let mut buf = vec![0u8; udp_frame_size as usize];
             loop {
                 match udp_sock.recv_from(&mut buf).await {
                     Ok((len, addr)) => {
                         trace!("udp packet {} bytes from {}", len, addr);
-                        if len > UDP_BUF_LEN {
-                            error!("udp packet from {} is too large ({})", addr, len);
-                        } else {
-                            let frame: Vec<u8> = buf[..len].iter().copied().collect();
-                            let ack_code = match process_udp_packet(frame).await {
-                                Ok(true) => Some(RESPONSE_OK),
-                                Err((e, need_ack)) => {
-                                    error!("udp packet from {} error: {}", addr, e);
-                                    if need_ack {
-                                        if e.kind() == psrt::ErrorKind::AccessDenied {
-                                            Some(RESPONSE_ERR_ACCESS)
-                                        } else {
-                                            Some(RESPONSE_ERR)
-                                        }
+                        let frame: Vec<u8> = buf[..len].iter().copied().collect();
+                        let ack_code = match process_udp_packet(frame).await {
+                            Ok(true) => Some(RESPONSE_OK),
+                            Err((e, need_ack)) => {
+                                error!("udp packet from {} error: {}", addr, e);
+                                if need_ack {
+                                    if e.kind() == psrt::ErrorKind::AccessDenied {
+                                        Some(RESPONSE_ERR_ACCESS)
                                     } else {
-                                        None
+                                        Some(RESPONSE_ERR)
                                     }
+                                } else {
+                                    None
                                 }
-                                _ => None,
-                            };
-                            if let Some(code) = ack_code {
-                                let mut buf = CONTROL_HEADER.to_vec();
-                                buf.extend(&psrt::PROTOCOL_VERSION.to_le_bytes());
-                                buf.push(code);
-                                if let Err(e) = udp_sock.send_to(&buf, addr).await {
-                                    error!("{}", e);
-                                }
+                            }
+                            _ => None,
+                        };
+                        if let Some(code) = ack_code {
+                            let mut buf = CONTROL_HEADER.to_vec();
+                            buf.extend(&psrt::PROTOCOL_VERSION.to_le_bytes());
+                            buf.push(code);
+                            if let Err(e) = udp_sock.send_to(&buf, addr).await {
+                                error!("{}", e);
                             }
                         }
                     }
@@ -973,16 +969,8 @@ async fn process_udp_packet(frame: Vec<u8>) -> Result<bool, (Error, bool)> {
         if !ALLOW_ANONYMOUS.load(atomic::Ordering::SeqCst) {
             return Err((Error::access("anonymous access failed"), need_ack));
         }
-    } else {
-        match authenticate(login, password).await {
-            Ok(false) => {
-                return Err((Error::access("authentication failed"), need_ack));
-            }
-            Err(e) => {
-                return Err((e, need_ack));
-            }
-            _ => {}
-        }
+    } else if !authenticate(login, password).await {
+        return Err((Error::access("authentication failed"), need_ack));
     }
     let acl = if let Some(acl) = get_acl(login).await {
         acl
@@ -1196,12 +1184,6 @@ fn main() {
     MAX_TOPIC_LENGTH.store(config.server.max_topic_length, atomic::Ordering::SeqCst);
     MAX_PUB_SIZE.store(config.server.max_pub_size, atomic::Ordering::SeqCst);
     psrt::set_max_topic_depth(config.server.max_topic_depth);
-    if let Some(ref f) = config.auth.password_file {
-        PASSWORD_FILE
-            .try_write()
-            .unwrap()
-            .replace(format_path!(f.clone()));
-    }
     if opts.daemonize {
         if let Ok(fork::Fork::Child) = fork::daemon(true, false) {
             std::process::exit(0);
@@ -1212,12 +1194,25 @@ fn main() {
         .enable_all()
         .build()
         .unwrap();
+    macro_rules! reload_passwords {
+        ($db: expr) => {
+            if let Err(e) = $db.reload().await {
+                error!("Unable to load password file: {}", e);
+            }
+        };
+    }
     rt.block_on(async move {
         {
-            let mut acl = ACL_DB.try_write().unwrap();
+            let mut acl = ACL_DB.write().await;
             acl.set_path(&format_path!(config.auth.acl));
             acl.reload().await.unwrap();
         }
+        if let Some(ref f) = config.auth.password_file {
+            let mut passwords = PASSWORD_DB.write().await;
+            let pfile = format_path!(f.clone());
+            passwords.set_password_file(&pfile);
+            reload_passwords!(passwords);
+        }
         handle_term_signal!(SignalKind::interrupt(), false);
         handle_term_signal!(SignalKind::terminate(), true);
         tokio::spawn(async move {
@@ -1238,6 +1233,9 @@ fn main() {
                     if let Err(e) = acl_dbm!().reload().await {
                         error!("ACL reload failed: {}", e);
                     }
+                    {
+                        reload_passwords!(PASSWORD_DB.write().await);
+                    }
                 }
             }
         });
@@ -1376,18 +1374,12 @@ mod stats {
             };
             let mut authorized = false;
             if let Some((ref login, ref password)) = credentials {
-                match super::authenticate(login, password).await {
-                    Ok(true) => {
-                        if let Some(acl) = super::get_acl(login).await {
-                            if acl.is_admin() {
-                                authorized = true;
-                            }
+                if super::authenticate(login, password).await {
+                    if let Some(acl) = super::get_acl(login).await {
+                        if acl.is_admin() {
+                            authorized = true;
                         }
                     }
-                    Err(e) => {
-                        return response!(StatusCode::INTERNAL_SERVER_ERROR, e.to_string());
-                    }
-                    _ => {}
                 }
             } else if let Some(acl) = super::get_acl("_").await {
                 if acl.is_admin() {
